//
// A simple example of OpenGL usage over X11:
// draw a regulat tetrahedron inside a cube
//
// The 3D model can be rotated by the mouse movement
// with the left mouse button pressed.
//
// The program is based on the class GLWindow, that presents a
// convenient C++ interface to OpenGL and X11 functions.
//
//                                      Written by V.Borisenko
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "GLWindow.h"

//--------------------------------------------------
// Definition of class "MyWindow"
//
class MyWindow: public GLWindow {  // Our main class derived from GLWindow
    GLfloat         m_Alpha;    // Angle of rotation around vert.axis in degrees
    GLfloat         m_Beta;     // Angle of rotation around hor.axis in degrees
    I2Point         m_MousePos; // Previous position of mouse pointer
public:
    MyWindow():                 // Constructor
        GLWindow(),
        m_Alpha(130.),
        m_Beta(196.),
        m_MousePos(-1, -1)
    {}

    void render();              // Draw a scene graph

    virtual void onExpose(XEvent& event);
    virtual void onKeyPress(XEvent& event);
    virtual void onButtonPress(XEvent& event);
    virtual void onButtonRelease(XEvent& event);
    virtual void onMotionNotify(XEvent& event);
};

//
// Process the Expose event: draw in the window
//
void MyWindow::onExpose(XEvent& /* event */) {
    makeCurrent();

    // Rotate the scene
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glEnable(GL_LIGHT0);        // Turn on the first light source

    // Rotate a model
    glRotatef(m_Beta, 1., 0., 0.);
    glRotatef(m_Alpha, 0., 1., 0.);

    render();           // Draw a scene graph

    swapBuffers();
}

//
// Process the KeyPress event: 
// if "q" is pressed, then close the window
//
void MyWindow::onKeyPress(XEvent& event) {
    KeySym key;
    char keyName[256];
    int nameLen = XLookupString(&(event.xkey), keyName, 255, &key, 0);
    printf("KeyPress: keycode=0x%x, state=0x%x, KeySym=0x%x\n",
        event.xkey.keycode, event.xkey.state, (int) key);
    if (nameLen > 0) {
        keyName[nameLen] = 0;
        printf("\"%s\" button pressed.\n", keyName);
        if (keyName[0] == 'q') { // quit => close window
            destroyWindow();
        }
    }
}

// Process mouse click
void MyWindow::onButtonPress(XEvent& /* event */) {
    // int x = event.xbutton.x;
    // int y = event.xbutton.y;
    // int mouseButton = event.xbutton.button;

    // printf("Mouse click: x=%d, y=%d, button=%d\n", x, y, mouseButton);

    m_MousePos.x = (-1); m_MousePos.y = (-1); // Mouse position undefined
}

void MyWindow::onButtonRelease(XEvent& /* event */) {
    m_MousePos.x = (-1); m_MousePos.y = (-1); // Mouse position undefined
}

//
// Process the mouse movement
//
void MyWindow::onMotionNotify(XEvent& event) {
    int x = event.xbutton.x;
    int y = event.xbutton.y;
    if ((event.xbutton.state & Button1Mask) != 0) {
        if (m_MousePos.x >= 0) {
            int dx = x - m_MousePos.x;
            int dy = y - m_MousePos.y;
            if (m_Beta > 100. || m_Beta < (-100.))
                dx = -dx;

            m_Alpha += (GLfloat) dx * 0.1;
            if (fabs(m_Alpha) > 360.)
                m_Alpha -= ((int) m_Alpha / 360) * 360.;

            m_Beta += (GLfloat) dy * 0.1;
            if (fabs(m_Beta) > 360.)
                m_Beta -= ((int) m_Beta / 360) * 360.;

            // printf("alpha=%f, beta=%f\n", m_Alpha, m_Beta);

            redraw();
        }
        m_MousePos.x = x; m_MousePos.y = y;
    } else {
        m_MousePos.x = (-1); m_MousePos.y = (-1);   // Undefined
    }
}

//
// Draw a 3D model
// Cube vertices:
//      7-----6   Tetrahedron vertices:
//    / |    /|   0 2 5 7
//   4------5 |
//   |  |   | |
//   |  3 - |-2
//   |/     |/
//   0------1
static const GLfloat cube_vertices[8][3] = {
    { -1., -1., -1. },
    {  1., -1., -1. },
    {  1.,  1., -1. },
    { -1.,  1., -1. },

    { -1., -1.,  1. },
    {  1., -1.,  1. },
    {  1.,  1.,  1. },
    { -1.,  1.,  1. }
};

void MyWindow::render() {
    GLfloat color[4], normal[3];

    glClearColor(0.2, 0.3, 0.5, 1.); // Background color: dark blue
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // Draw a tetrahedron
    glBegin(GL_TRIANGLES);
        // Face 0 2 5 ----------------
        color[0] = 0.2; color[1] = 1.; color[2] = 0.2;  // Green
        glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);

        // Set a normal to the face 0 2 5
        normal[0] = 1.;
        normal[1] = -1.;
        normal[2] = -1.;
        glNormal3fv(normal);

        // Define vertices of the face
        glVertex3fv(cube_vertices[0]);
        glVertex3fv(cube_vertices[2]);
        glVertex3fv(cube_vertices[5]);

        // Face 0 7 2 ----------------
        color[0] = 1.0; color[1] = 1.0; color[2] = 0.2;  // Yellow
        glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);

        // Set a normal to the face 0 7 2
        normal[0] = -1.;
        normal[1] = 1.;
        normal[2] = -1.;
        glNormal3fv(normal);

        // Define vertices of the face
        glVertex3fv(cube_vertices[0]);
        glVertex3fv(cube_vertices[7]);
        glVertex3fv(cube_vertices[2]);

        // Face 2 7 5 ----------------
        color[0] = 1.0; color[1] = 0.2; color[2] = 0.2; // Red
        glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);

        // Set a normal to the face 2 7 5
        normal[0] = 1.;
        normal[1] = 1.;
        normal[2] = 1.;
        glNormal3fv(normal);

        // Define vertices of the face
        glVertex3fv(cube_vertices[2]);
        glVertex3fv(cube_vertices[7]);
        glVertex3fv(cube_vertices[5]);

        // Face 0 5 7 ----------------
        color[0] = 0.2; color[1] = 0.2; color[2] = 1.0;  // Blue
        glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);

        // Set a normal to the face 2 7 5
        normal[0] = -1.;
        normal[1] = -1.;
        normal[2] = 1.;
        glNormal3fv(normal);

        // Define vertices of the face
        glVertex3fv(cube_vertices[0]);
        glVertex3fv(cube_vertices[5]);
        glVertex3fv(cube_vertices[7]);
    glEnd();

    color[0] = 1.0; color[1] = 1.0; color[3] = 1.0; // White
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
    glBegin(GL_LINES);
        // Edge 0 1 -----------
        normal[0] = 0.;
        normal[1] = -1.;
        normal[2] = -1.;
        glNormal3fv(normal);

        glVertex3fv(cube_vertices[0]);
        glVertex3fv(cube_vertices[1]);

        // Edge 1 2 -----------
        normal[0] = 1.;
        normal[1] = 0.;
        normal[2] = -1.;
        glNormal3fv(normal);

        glVertex3fv(cube_vertices[1]);
        glVertex3fv(cube_vertices[2]);

        // Edge 2 3 -----------
        normal[0] = 0.;
        normal[1] = 1.;
        normal[2] = -1.;
        glNormal3fv(normal);

        glVertex3fv(cube_vertices[2]);
        glVertex3fv(cube_vertices[3]);

        // Edge 3 0 -----------
        normal[0] = -1.;
        normal[1] = 0.;
        normal[2] = -1.;
        glNormal3fv(normal);

        glVertex3fv(cube_vertices[3]);
        glVertex3fv(cube_vertices[0]);

        // Edge 4 5 -----------
        normal[0] = 0.;
        normal[1] = -1.;
        normal[2] = 1.;
        glNormal3fv(normal);

        glVertex3fv(cube_vertices[4]);
        glVertex3fv(cube_vertices[5]);

        // Edge 5 6 -----------
        normal[0] = 1.;
        normal[1] = 0.;
        normal[2] = 1.;
        glNormal3fv(normal);

        glVertex3fv(cube_vertices[5]);
        glVertex3fv(cube_vertices[6]);

        // Edge 6 7 -----------
        normal[0] = 0.;
        normal[1] = 1.;
        normal[2] = 1.;
        glNormal3fv(normal);

        glVertex3fv(cube_vertices[6]);
        glVertex3fv(cube_vertices[7]);

        // Edge 7 0 -----------
        normal[0] = -1.;
        normal[1] = 0.;
        normal[2] = 1.;
        glNormal3fv(normal);

        glVertex3fv(cube_vertices[7]);
        glVertex3fv(cube_vertices[4]);

        // Edge 0 4 -----------
        normal[0] = -1.;
        normal[1] = -1.;
        normal[2] = 0.;
        glNormal3fv(normal);

        glVertex3fv(cube_vertices[0]);
        glVertex3fv(cube_vertices[4]);

        // Edge 1 5 -----------
        normal[0] = 1.;
        normal[1] = -1.;
        normal[2] = 0.;
        glNormal3fv(normal);

        glVertex3fv(cube_vertices[1]);
        glVertex3fv(cube_vertices[5]);

        // Edge 2 6 -----------
        normal[0] = 1.;
        normal[1] = 1.;
        normal[2] = 0.;
        glNormal3fv(normal);

        glVertex3fv(cube_vertices[2]);
        glVertex3fv(cube_vertices[6]);

        // Edge 3 7 -----------
        normal[0] = -1.;
        normal[1] = 1.;
        normal[2] = 0.;
        glNormal3fv(normal);

        glVertex3fv(cube_vertices[3]);
        glVertex3fv(cube_vertices[7]);

    glEnd();
}

/////////////////////////////////////////////////////////////
// Main: initialize X, create an instance of MyWindow class,
//       and start the message loop
int main() {
    // Initialize X stuff
    if (!GWindow::initX()) {
        printf("Could not connect to X-server.\n");
        exit(1);
    }

    int width = GWindow::screenMaxX()/2;
    int height = GWindow::screenMaxY()/2;
    double aspect = (double) width / (double) height;

    MyWindow w;
    w.createWindow(
        I2Rectangle(                    // Window frame rectangle:
            I2Point(10, 10),            //     left-top corner,
            width, height               //     width, height
        ),
        R2Rectangle(                    // Window coordinate rectangle
            R2Point(-2. * aspect, -2.), //     left-top corner,
            4. * aspect, 4.             //     width, height
        ),
        "Tetrahedron in Cube"           // Window title
    );
    w.setBackground("lightGray");
    w.makeCurrent();
    GLWindow::initializeOpenGL();

    GWindow::messageLoop();

    GWindow::closeX();
    return 0;
}
